/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "project.h"
  
int symbol::nullsymbolcount = 0;

symbol::symbol(int name, attribute *attributes, symbol *parent)
{
  symbol::name = name;
  symbol::attributes = attributes;
  symbol::parent = parent;
  symbol::children = (list<symbol> *) NULL;
  if (parent != NULL) parent->children = new list<symbol>(this, parent->children);
}

symbolname *symbol::getfullname()
{
  if (parent == NULL) return new list<int>(new int(name), (list<int> *) NULL);
  else return  new list<int>(new int(name), parent->getfullname());
}

char *symbol::getfullnametext()
{
  static char buff[2][100];
  static int _curbuff = 0;
  _curbuff ^= 1;
  int curbuff = _curbuff;

  if (parent == NULL) return symbtable.tostring(name);
  snprintf(buff[curbuff], 100, "%s/%s", parent->getfullnametext(),symbtable.tostring(name));
  return buff[curbuff];
}

symbolwidget *symbol::getwidget()
{
  return new plainsymbolwidget(attributes, this);
}

symbol *symbol::getlocalsymbol(symbolname *s)
{
  if (s == NULL) return this;

  for (list<symbol> *l = children; l != NULL; l = l->next)
    if (l->elem->getname() == *(s->elem)) 
      return l->elem->getlocalsymbol(s->next);  

  return _getlocalsymbol(s);
}

symbol *symbol::getparent(symbolname *s)
{
  if (s == NULL || s->next == NULL) return (symbol *) NULL;

  symbolname *sn;
  for (sn = s; sn->next->next != NULL; sn = sn->next);

  symbolname *temp = sn->next; // Bad cludge. I hate it!!
  sn->next = (symbolname *) NULL;
  symbol *symb = getsymbol(sn);
  sn->next = temp;

  return symb;
}

symbol *symbol::getsymbol(char *s)
{
  list<int> *n = symbtable.toname(s);
  if (n == NULL) return (symbol *) NULL;
  
  symbol *sy = getsymbol(n);
  symbtable.deletename(n);
  return sy;
}

symbolconstructor::symbolconstructor(int name, attribute *attribs) 
{ 
  fullyinitialized = 0;
  defline = GetParseLineNum(); 
  DefFile = GetParseFileName();
  symbolconstructor::attribs = nullattribute;
  initialize(name, attribs);
}

symbolconstructor::symbolconstructor(attribute *paddingattribute)
{
  fullyinitialized = 0;
  defline = GetParseLineNum(); 
  DefFile = GetParseFileName();
  attribs = paddingattribute;
}

symbol *symbolconstructor::_create(symbol *parent) 
{ 
  createerror(TYPE_INTERNALERROR, "symbolconstructor::create called on %s", symbtable.tostring(name)); 

  return (symbol *) NULL; // This should never happen as createerror exits.
}

void symbolconstructor::addattribstoend(attribute *att) 
{ 
  attribs = attribs->addattribstoend(att); 
}

void symbolconstructor::addattribs(attribute *att) 
{ 
  attribs = att->addattribstoend(attribs); 
}

void symbolconstructor::ccreate(macrocomponent *parent)
{
  if (!fullyinitialized) 
  {
    createerror(TYPE_INTERNALERROR, "Creation of unfullyinitialized constructor %s. Skipped.", symbtable.tostring(name)); 
    return;
  }
  _ccreate(parent);
}

symbol * symbolconstructor::_ccreate(macrocomponent *parent)
{
  // This symbol is not created at component creation time.

  return (symbol *) NULL;
}

symbol *symbolconstructor::create(symbol *parent) 
{ 
  if (!fullyinitialized) 
  {
    createerror(TYPE_INTERNALERROR, "Creation of unfullyinitialized constructor %s. Skipped...", symbtable.tostring(name)); 
    return (symbol *) NULL;
  }
  return _create(parent); 
}

void symbolconstructor::initialize(int name, attribute *attribs)
{
  if (fullyinitialized) createerror(TYPE_INTERNALERROR, "%s symbolconstructor initialized twice.", symbtable.tostring(name));
  symbolconstructor::attribs = attribs->addattribstoend(symbolconstructor::attribs); /*12345*/
  symbolconstructor::name = name; 
  fullyinitialized = 1;
}

void symbolconstructor::createerror(int type, char *Msg, ...)
{ 
  va_list val;
  va_start(val, Msg);
  
  vaerror(type, defline, DefFile, Msg, val);
}

symbol *nullsymbol(attribute *attribs, symbol *parent)
{
  char s[] = "__nullsymbol__xxxxxxxxxx";
  sprintf(s, "__nullsymbol%i__", symbol::nullsymbolcount);
  symbol::nullsymbolcount++;
  return new symbol(symbtable.toid(s), attribs, parent);
}

symbol *symbol::getsymbol(symbolname *s)
{
  symbol *sym = getlocalsymbol(s);
  if (sym != NULL) return sym;
  return Project.getlocalsymbol(s);
}

int symbol::getname()
{
  return name; 
}

int symbolconstructor::getname()
{
  return name; 
}

attribute *symbol::getattribute(int name)
{
  attribute *attr = attributes->getattribute(name);
  if (attr != NULL) return attr;
  if (parent == NULL) return nullattribute;
  attr = parent->getattribute(name);
  if (attr != NULL) return attr;
  return nullattribute;
}

symbolcopyconstructor::symbolcopyconstructor(int name, attribute *attribs, symbolname *orig) : symbolconstructor(name, attribs)
{
  symbolcopyconstructor::orig = orig;
}

symbol *symbolcopyconstructor::_create(symbol *parent)
{
  if (parent == NULL) 
  {
    createerror(TYPE_INTERNALERROR, "Tried to create symbolcopy with NULL parent.");
    return NULL;
  }

  symbol *s = parent->getsymbol(orig);
  if (s == NULL) 
  {
    createerror(TYPE_ERROR, "Can't find symbol %s to create a mirror of it.", symbtable.tostring(orig));
    return NULL;
  }

  return new symbolptr(name, attribs, parent, s);
}

symbolptr::symbolptr(int name, attribute *attribs, symbol *parent, symbol *s) : symbol(name, attribs, parent) 
{ 
  ptr = s; 
  nameattrib = symbtable.isanonymous(name) ? nullattribute : stringattribute(symbtable.toid("disptitle"), symbtable.tostring(name)); 
}
